{% extends "!layout.html" %}
{% block rootrellink %}
   <li>[<a href="{{ pathto('installation') }}">install</a></li>
   <li>|<a href="{{ pathto('overview') }}">overview</a></li>
   <li>|<a href="{{ pathto('pv') }}">pv</a></li>
   <li>|<a href="{{ pathto('ca') }}">ca</a></li>
   <li>|<a href="{{ pathto('arrays') }}">arrays</a></li>
   <li>|<a href="{{ pathto('devices') }}">devices</a></li>
   <li>|<a href="{{ pathto('alarm') }}">alarm</a></li>
   <li>|<a href="{{ pathto('autosave') }}">autosave</a></li>
   <li>|<a href="{{ pathto('wx') }}">wx</a></li>
   <li>|<a href="{{ pathto('advanced') }}">advanced</a>]</li>
{% endblock %}

{% block relbar1 %}
<div style="background-color: #E7F0F3; text-align: left; padding: 8px 4px 8px 4px">
  <font size=+3><a href="{{ pathto('index') }}">
      <img src="{{pathto("_static/pyepics.png", 1) }}" height=50 border="0" alt="pyepics"/>
    &nbsp; PyEpics: Epics Channel Access for Python</a>
  </font>
</div>
{{ super() }}
{% endblock %}


{# put the sidebar before the body #}
{% block sidebar1 %}{{ sidebar() }}{% endblock %}
{% block sidebar2 %}{% endblock %}
