<h3>Download</h3>
<p>Current version: <b>{{ release }}</b></p>


<p>
  Install: &nbsp; <tt> pip install pyepics</tt>

<p>
  Develop: &nbsp; <a href="https://github.com/pyepics/pyepics/">github.com</a>
<br>

<p>
